/* 
 * Linux test app for https://www.virtualbox.org/ticket/17961
 *
 *     gcc vmdetect.c -std=gnu11 -masm=intel -o vmdetect
 *
 */

#define _POSIX_C_SOURCE 200809L

#include <signal.h>
#include <stdio.h>

#if __x86_64__
#define ENABLE_TF() asm( \
        "pushf;" \
        "or WORD ptr[rsp], 0x100;" \
        "popf;" \
        );

#define DISABLE_TF() asm( \
        "pushf;" \
        "and WORD ptr[rsp], 0xFFEF;" \
        "popf;" \
        );
#else
#define ENABLE_TF() asm( \
        "pushf;" \
        "or WORD ptr[esp], 0x100;" \
        "popf;" \
        );

#define DISABLE_TF() asm( \
        "pushf;" \
        "and WORD ptr[esp], 0xFFEF;" \
        "popf;" \
        );
#endif

static int codepoint = -1;

static void handler(int signo, siginfo_t *info, void *context) {
    // store codepoint of first exception
    if (codepoint < 0) {
        codepoint = *(unsigned char*)info->si_addr;
    }
}


struct sigaction handler_act = {
    .sa_sigaction = &handler,
    .sa_flags = SA_SIGINFO,
};

void cpuid_test()
{
    ENABLE_TF();

    asm(
        "cpuid;"
        "nop;"        // normal TF
        "int 3;"      // in VM
        : /* no output */
        : /* no input */
        : "eax", "ebx", "ecx", "edx" /* clobber list */
    );

    DISABLE_TF();
}

int main(int argc, char *argv[]) {
    if (sigaction(SIGTRAP, &handler_act, NULL) < 0) {
        printf("Error installing signal handler");
		return 1;
	}

    cpuid_test();

    printf("Found opcode 0x%02hhX\n", codepoint);
    if (codepoint == 0x90) {
        printf("Trapped on nop, we are running on a real machine\n");
    } else if (codepoint == 0xCC) {
        printf("Trapped on int3, we are running in a VM\n");
    } else {
        printf("Unexpected trap location.\n");
    }

    return 0;
}

